-- This file:
--   http://angg.twu.net/AGDA/Hu1.agda.html
--   http://angg.twu.net/AGDA/Hu1.agda
--           (find-angg "AGDA/Hu1.agda")
-- Author: Eduardo Ochs <eduardoochs@gmail.com>
-- Version: 20220205
--
-- Based on:
-- (find-es "agda" "category-theory")
-- (find-agdacatsfile "src/Categories/Category/Core.agda")
-- (find-agdacatsfile "src/Categories/Functor/Core.agda")
-- (find-agdacatsfile "src/Categories/NaturalTransformation/Core.agda")
--
-- (defun a () (interactive) (find-angg "AGDA/Hu1.agda"))

module Hu1 where

open import Level
open import Function.Base using (flip)
open import Relation.Binary using (Rel; IsEquivalence; Setoid)
open import Relation.Binary.PropositionalEquality as ≡ using (_≡_)
import Relation.Binary.Reasoning.Setoid as SetoidR

private
  variable
    o ℓ e o′ ℓ′ e′ : Level

record Category (o ℓ e : Level) : Set (suc (o ⊔ ℓ ⊔ e)) where
  eta-equality
  infix  4 _⇒_
  infixr 9 _∘_

  field
    Obj : Set o
    _⇒_ : Rel Obj zero
    id  : ∀ {A} → (A ⇒ A)
    _∘_ : ∀ {A B C} → (B ⇒ C) → (A ⇒ B) → (A ⇒ C)

  field
    assoc     : ∀ {A B C D} {f : A ⇒ B} {g : B ⇒ C} {h : C ⇒ D} → (h ∘ g) ∘ f ≡ h ∘ (g ∘ f)
    sym-assoc : ∀ {A B C D} {f : A ⇒ B} {g : B ⇒ C} {h : C ⇒ D} → h ∘ (g ∘ f) ≡ (h ∘ g) ∘ f
    identityˡ : ∀ {A B} {f : A ⇒ B} → id ∘ f ≡ f
    identityʳ : ∀ {A B} {f : A ⇒ B} → f ∘ id ≡ f
    identity² : ∀ {A} → id ∘ id {A} ≡ id {A}

  o-level : Level
  o-level = o
  ℓ-level : Level
  ℓ-level = ℓ
  e-level : Level
  e-level = e

-- Convenience functions for working over mupliple categories at once:
-- C [ x , y ] (for x y objects of C) - Hom_C(x , y)
-- C [ f ≈ g ] (for f g arrows of C)  - that f and g are equivalent arrows
-- C [ f ∘ g ] (for f g composables arrows of C) - composition in C
infix 10  _[_,_] _[_≈_] _[_∘_]

_[_,_] : (C : Category o ℓ e) → (X : Category.Obj C) → (Y : Category.Obj C) → Set
_[_,_] = Category._⇒_

_[_≈_] : (C : Category o ℓ e) → ∀ {X Y} (f g : C [ X , Y ]) → Set
_[_≈_] C {X} {Y} f g = f ≡ g

_[_∘_] : (C : Category o ℓ e) → ∀ {X Y Z} (f : C [ Y , Z ]) → (g : C [ X , Y ]) → C [ X , Z ]
_[_∘_] = Category._∘_


-- Combinators for commutative diagram
-- The idea is to use the combinators to write commutations in a more readable way.
-- It starts with [_⇒_]⟨_≈_⟩, and within the third and fourth places, use _⇒⟨_⟩_ to
-- connect morphisms with the intermediate object specified.
module Commutation (𝓒 : Category o ℓ e) where
  open Category 𝓒

  infix 1 [_⇒_]⟨_≈_⟩
  [_⇒_]⟨_≈_⟩ : ∀ (A B : Obj) → A ⇒ B → A ⇒ B → Set _
  [ A ⇒ B ]⟨ f ≈ g ⟩ = f ≡ g

  -- infixl 2 connect
  connect : ∀ {A C : Obj} (B : Obj) → A ⇒ B → B ⇒ C → A ⇒ C
  connect B f g = g ∘ f

  syntax connect B f g = f ⇒⟨ B ⟩ g


record Functor (C : Category o ℓ e) (D : Category o′ ℓ′ e′) : Set (o ⊔ ℓ ⊔ e ⊔ o′ ⊔ ℓ′ ⊔ e′) where
  eta-equality
  private module C = Category C
  private module D = Category D

  field
    F₀ : C.Obj → D.Obj
    F₁ : ∀ {A B} (f : C [ A , B ]) → D [ F₀ A , F₀ B ]

    identity     : ∀ {A} → D [ F₁ (C.id {A}) ≈ D.id ]
    homomorphism : ∀ {X Y Z} {f : C [ X , Y ]} {g : C [ Y , Z ]} →
                     D [ F₁ (C [ g ∘ f ]) ≈ D [ F₁ g ∘ F₁ f ] ]
    F-resp-≈     : ∀ {A B} {f g : C [ A , B ]} → C [ f ≈ g ] → D [ F₁ f ≈ F₁ g ]

  -- nice shorthands
  ₀ = F₀
  ₁ = F₁





module NT where
  open import Level
  open Category
  open Functor -- renaming (id to idF)
  
  private
    variable
      -- o ℓ e o′ ℓ′ e′ : Level
      C D E : Category o ℓ e
  
  record NaturalTransformation {C : Category o ℓ e}
                               {D : Category o′ ℓ′ e′}
                               (F G : Functor C D) : Set (o ⊔ ℓ ⊔ ℓ′ ⊔ e′) where
    eta-equality
    private
      module F = Functor F
      module G = Functor G
      module C = Category C
      module D = Category D
    open F using (F₀; F₁)
    open Category D -- hiding (op)
  
    field
      η           : ∀ X → D [ F.F₀ X , G.F₀ X ]
      commute     : ∀ {X Y} (f : C [ X , Y ]) → η Y D.∘ F.F₁ f ≡ G.F₁ f D.∘ η X
      -- We introduce an extra proof to ensure the opposite of the opposite of a natural
      -- transformation is definitionally equal to itself.
      sym-commute : ∀ {X Y} (f : C [ X , Y ]) → G.F₁ f D.∘ η X ≡ η Y D.∘ F.F₁ f




-- Local Variables:
-- coding:  utf-8-unix
-- End:
