;;; eev-gud.el --- functions to place debugger hyperlinks in e-scripts.

;; Copyright (C) 2000, 2001, 2002 Eduardo Ochs.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2002may11
;; Keywords:   gdb, gud, debuggers, eev
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;;
;; URLs for this file:
;;   http://angg.twu.net/eev/eev-gud.el
;;   http://angg.twu.net/eev/eev-gud.el.html

;;; Commentary:

;; This file defines gdbk-mode, that adds some convenient keybindings
;; on top of GUD, the gdbk-xxxdb family of functions, for hyperlinks
;; that activate debuggers, and also ee-comint-send-line, that sends
;; individual lines to a GUD buffer and thus works as a sort of an
;; Emacs-based eeg.
;;
;; A demo: (find-es "anatocc" "dlopen")

;; (find-elnode "Autoload" "magic autoload comment")
;; (find-elnode "Comment Tips")
;; (find-elnode "Library Headers")
;; (find-elnode "Documentation Tips")
;; (find-elnode "Documentation Tips" "single-quotes")
;; (find-elnode "Keys in Documentation")
;; (find-es "escripts" "eev-tutorial")

;;; History:

;; 2001mar20: took the code from eev-extras.el and made a separate
;;            file.

;;; Index:

;; .eegud-keys-mode		(to "eegud-keys-mode")
;; .eegud-hyperlinks		(to "eegud-hyperlinks")
;; .ee-comint-send-line	(to "ee-comint-send-line")
;; .ee-read-eeg-lines		(to "ee-read-eeg-lines")



;;; Code:

;; eegud-keys-mode  (to ".eegud-keys-mode")
;; A minor mode with some key redefinitions to make GUD easier to
;; use; all my "eegud-hyperlinks" use this mode, that makes some of
;; the keys described in this page work:
;;   (find-node "(gdb)Emacs")
;; Plus, M-L and M-R will send individual lines from $EEG to the
;; GUD buffer (but I'm using the global keymap to bind them) and
;; M-k from the GUD buffer will kill the buffer and its window.
;;
;; (find-es "emacs" "easy-mmode")
;; (find-es "emacs" "meta-shift-letter")
;; (find-elnode "Function Cells")
;; (find-elnode "Easy-Mmode")
;; (find-e20file "emacs-lisp/easy-mmode.el")
;; (find-e20file "emacs-lisp/pp.el")
;; (find-elnode "Mode Line Variables" "Variable: minor-mode-alist")
;; (insert (pp (symbol-function 'gdbk-mode)))
;; (insert (pp minor-mode-alist))
;; (find-etag "kill-buffer-and-window")

;; (find-e20node "Starting GUD")
;; (find-e20node "Debuggers")
;; (find-e20node "Basic Window")
;; (find-e20node "Commands of GUD")
;; (find-e20node "GUD Customization")

;; XEmacs doesn't have easy-mmode, AFAIK, so...
;; (Note: the condition used to be "(if-gnu-emacs ...)").
;; I have used XEmacs very little, though.

(if (fboundp 'easy-mmode-define-minor-mode)

(easy-mmode-define-minor-mode
 eegud-keys-mode
 "eegud keys mode.\nUse `C-c C-t' and `C-c C-d' to set and delete breakpoints.\n
See: (find-e20node \"Commands of GUD\")"
 nil
 " eegudk"
 '(("\M-s" . gud-step)
   ("\M-n" . gud-next)
   ("\M-c" . gud-cont)
   ("\M-f" . gud-finish)
   ("\M-k" . (lambda (s) (interactive "sConfirm killing: ")
	       (let ((buffer (current-buffer)))
		 (condition-case nil (delete-window (selected-window)))
		 ;; (noerr (delete-window (selected-window)))
		 (kill-buffer buffer))))))

) ; end of the "if (fboundp ...)" block.

;; eegud-hyperlinks  (to ".eegud-hyperlinks")
;; (find-es "debian" "sudo-gdbk")
;; Grep my ".e"s for "gdbk" to take a look at the old way.

(defun eegud-gdb (do-eeg-bounded dir &optional rest load-eeg)
  (if do-eeg-bounded (eegdb-bounded (if ee-arg nil 'once)))
  (gdb (format "gdb %s%s" dir (or rest "")))
  (if load-eeg (ee-read-eeg-lines))
  (eegud-keys-mode))

(defun eegud-pdb (do-eeg-bounded dir &optional rest load-eeg)
  (if do-eeg-bounded (eeg-bounded))
  (pdb (format "/usr/lib/python1.5/pdb.py %s%s" dir (or rest "")))
  (if load-eeg (ee-read-eeg-lines))
  (eegud-keys-mode))

(defun eegud-perldb (do-eeg-bounded dir &optional rest load-eeg)
  (if do-eeg-bounded (eeg-bounded))
  (perldb (format "perl %s%s" dir (or rest "")))
  (if load-eeg (ee-read-eeg-lines))
  (eegud-keys-mode))


;; ee-comint-send-line  (to ".ee-comint-send-line")
;; ee-read-eeg-lines  (to ".ee-read-eeg-lines")
;; (find-es "emacs" "meta-shift-letter")

(global-set-key "\M-L" 'ee-comint-send-line)
(global-set-key "\M-R" 'ee-read-eeg-lines)

(setq ee-lines-list nil)

(defun ee-read-eeg-lines () (interactive)
  (setq ee-lines-list
	(with-temp-buffer
	  (insert-file-contents (substitute-in-file-name "$EEG"))
	  (let ((lines nil) beg end (max (point-max)))
	    (goto-char 0)
	    (while (< (setq beg (point)) max)
	      (end-of-line)
	      (setq lines (cons (buffer-substring beg (point)) lines))
	      (noerr (forward-char)))
	    (reverse lines)))))

(defun ee-comint-send-line () (interactive)
  (if ee-lines-list (progn (insert (car ee-lines-list))
			   (comint-send-input)
			   (setq ee-lines-list (cdr ee-lines-list)))))



;; Local Variables:
;; coding:               no-conversion
;; ee-comment-format:	 ";; %s\n"
;; ee-anchor-format:     "%s"
;; ee-charset-indicator: ""
;; End:
