;; Copyright (C) 1999, 2000, 2001, 2002 Free Software Foundation, Inc.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2002may11
;; Keywords:   escripts, hyperlinks, shell, tcl, perl, tex
;;
;; This file was copylefted to prevent against patent psychopaths; if
;; you want a version with any other license you'll have to write it
;; yourself. More formally,
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;;
;; URLs for this file:
;;   http://angg.twu.net/eev/eev-hybrids.el
;;   http://angg.twu.net/eev/eev-hybrids.el.html

;;; Commentary:

;; This package provides support for e-scripts in Emacs. Look in
;; <http://angg.twu.net/emacs.html> for more information and in
;; <http://angg.twu.net/escripts.html> for examples of e-scripts... an
;; e-script is typically a text file in which certain parts are meant
;; to be executed by the shell or by other programs and certain
;; comments are in fact Emacs Lisp code; executing one of these
;; comments with C-x C-e (i.e., eval-last-sexp) brings up a file or an
;; info node, and thus works as a hyperlink. Many of the comments in
;; this file can be executed in this way.

;; See also the "eev Manifesto", at
;; <http://angg.twu.net/eev-manifesto.html>, the section named
;; "eev-howto", below, and whatever else seem important in
;; <http://angg.twu.net/>; that site is almost completely dedicated to
;; using eev.el.

;;;;
;;;; eeman  (to ".eeman")

;;;; eeman: one way to link to specific points on manpages.
;;;; This is a low-level function; you should write your hyperlinks to
;;;; manpages using `find-man' instead of this one... (but take care,
;;;; `find-man' is not mature enough yet)
;;;;
;;
;; This function is a hybrid between the find-xxx's and the
;; eev's; it works as a link to a specific point on a manpage,
;; but you need to go to a shell and type "ee" to let it call man.

;;
;; Examples of usage:
;;
;; (eeman "man"  "  -P")
;; (eeman "less" "  \\+cmd")
;; (eeman "expect" 972)
;;
;;;###autoload
(defun eeman (manpage &optional pos-spec)
"Write a call to \"man\" into the temporary script file.
This function is used to make hyperlinks to manpages. For example,
if you place the point at the end of one of the three lines below\n
  # (eeman \"etags\")
  # (eeman \"1 man\" \"-P pager, --pager=pager\")
  # (eeman \"expect\" 972)\n
and type `\\[eval-last-sexp]', Emacs will write into the temporary script file
(see `eev') a call to \"man MANPAGE\", optionally adding some extra
arguments to it make it start from the line given by POS-SPEC; if
POS-SPEC is a string then \"man\" will be called as\n
  man -P \"less '+/POS-SPEC'\" MANPAGE\n
which means: \"use `less' to display the manpage, but tell it to search
for the regexp POS-SPEC before giving control to the user; if POS-SPEC
is a number then interpret it as a line number, and tell \"less\" to go
to the POS-SPECth line instead of searching for a string."
  (cond ((numberp pos-spec)
	 (eev (format "man -P 'less +%d' %s\n" pos-spec manpage) nil))
	((stringp pos-spec)
	 (eev (format "man -P \"less '+/%s'\" %s\n" pos-spec manpage) nil))
	(t
	 (eev (format "man %s\n" manpage) nil))))


(defun find-dvipage (fname n)
  (let ((command (format "xdvi +%d %s &" n fname)))
    (eev command nil)
    command))
(defun find-zdvipage (fname n)
  (let ((command (format "zxdvi %s +%d &" fname n)))
    (eev command nil)
    command))
(defun find-pspage (fname n)
  (let ((command (format "gv -page %d %s &" n fname)))
    (eev command nil)
    command))

(defun code-ps (code psfile)
  (format-and-eval
   "(defun find-%spage (n &rest comments) (find-pspage %S n))"
   code psfile))
(defun code-dvi (code dvifile)
  (format-and-eval
   "(defun find-%spage (n &rest comments) (find-dvipage %S n))"
   code dvifile))
(defun code-zdvi (code zdvifile)
  (format-and-eval
   "(defun find-%spage (n &rest comments) (find-zdvipage %S n))"
   code zdvifile))




;; find-man.el  (to ".find-man.el")
;; (find-elnode "Scope")
;; (find-elnode "Sentinels")
;; (find-efile "man.el" "set-process-sentinel")
;; (find-efile "man.el" "defun Man-bgproc-sentinel")
;;
;; Sentinel functions run at top-level, so we need a global
;; pos-spec-list (binding it in find-man.el wouldn't do), and we call
;; it ee-pos-spec-list-for-find-man.el. Always after using
;; ee-pos-spec-list-for-find-man.el we set it back to nil to avoid
;; letting it interfere with invoking man the standard way.

(setq ee-pos-spec-list-for-find-man.el nil)

(defun find-man.el (manpage &rest pos-spec-list)
  (setq ee-pos-spec-list-for-find-man.el pos-spec-list)
  (man manpage))

;; (find-evariable 'Man-notify-method)
;; (find-efile "man.el" "defun Man-notify-when-ready")
;; (find-eev "eev.el" "ee-invade-global-" "eev-install-aggressively")
;;
(defun ee-install-find-man.el-hack ()
  (load-library "man")
  (fset 'Man-notify-when-ready-orig (symbol-function 'Man-notify-when-ready))
  (defun Man-notify-when-ready (man-buffer)
    (switch-to-buffer man-buffer)
    (apply 'ee-goto-position ee-pos-spec-list-for-find-man.el)
    (setq pos-spec-man nil))
  )


;;; eev-hybrids.el ends here
