;;; eev-links.el --- embed elisp hyperlinks in plain text.
;; Copyright (C) 1999, 2000, 2001, 2002 Free Software Foundation, Inc.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2002may11
;; Keywords:   escripts, hyperlinks, shell, tcl, perl, tex
;;
;; This file was copylefted to prevent against patent psychopaths; if
;; you want a version with any other license you'll have to write it
;; yourself. More formally,
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;;
;; URLs for this file:
;;   http://angg.twu.net/eev/eev-links.el
;;   http://angg.twu.net/eev/eev-links.el.html

;;; Commentary:

;; This package provides support for e-scripts in Emacs. Look in
;; <http://angg.twu.net/emacs.html> for more information and in
;; <http://angg.twu.net/escripts.html> for examples of e-scripts... an
;; e-script is typically a text file in which certain parts are meant
;; to be executed by the shell or by other programs and certain
;; comments are in fact Emacs Lisp code; executing one of these
;; comments with C-x C-e (i.e., eval-last-sexp) brings up a file or an
;; info node, and thus works as a hyperlink. Many of the comments in
;; this file can be executed in this way.

;; See also the "eev Manifesto", at
;; <http://angg.twu.net/eev-manifesto.html>, the section named
;; "eev-howto", below, and whatever else seem important in
;; <http://angg.twu.net/>; that site is almost completely dedicated to
;; using eev.el.

;;;;
;;;; find-  (to ".find-")
;;;; Basic "find-..." functions.
;;;;
;;
;; Note that find-node and find-fline are the only functions that call
;; ee-goto-position directly, and they (and the calls to them from
;; code-c-d and find-node2) are already prepared to handle more than
;; one pos-spec argument. However, the simplistic version of
;; ee-goto-position below will scream if it receives more then one
;; argument.
;; There's an example of an enhanced ee-goto-position in my .emacs;
;; look for a link to it in <http://angg.twu.net/>.
;; (find-angg "eev-extras.el" "new_goto-position")

;;;###autoload
(defun ee-goto-position (&optional pos-spec &rest rest)
  "Go to a certain position in the current buffer is POS-SPEC is not nil.
If POS-SPEC is a string, search for its first occurence, starting from
the beginning of the accessible area; if it is a number, go to that
line; if POS-SPEC if nil then do nothing and leave point where it is.\n
See the documentation for the functions `find-fline' and `find-node'."
  (when pos-spec
    (goto-char (point-min))
    (cond ((numberp pos-spec)
	   (forward-line (1- pos-spec)))
	  ((stringp pos-spec)
	   (search-forward pos-spec))
	  (t (error "This is not a valid pos-spec: %S" pos-spec)))
    (if rest (ee-goto-rest rest))))

(defun ee-goto-rest (list)
  "Process any extra arguments given to `ee-goto-position'.
You are encouraged to redefine this function to make it more powerful."
  (cond ((stringp (car list))
	 (search-forward (car list))
	 (ee-goto-rest (cdr list)))))

(if (not (fboundp 'Info-goto-node))
    (autoload 'Info-goto-node "info"))

;;;###autoload
(defun find-node (nodestr &rest pos-spec-list)
  "Open an Info node, and optionally go to a certain position in it.
This function is often used as a hyperlink: if you place the point
at the end of a line like one of those below\n
  # (find-node \"(emacs-lisp-intro)Run a Program\")
  # (find-node \"(emacs-e20)Lisp Eval\" \"C-x C-e\")
  # (find-node \"(emacs)Lisp Eval\")\n
and type `\\[eval-last-sexp]' Emacs will open the corresponding Info node.\n
For more information on the format of POS-SPEC-LIST, see the
documentation for the function `ee-goto-position'; for ways to omit the
\"(info-file-name)\" part of NODESTR, see the documentation for the
function `code-c-d'."
  (if (Info-goto-node nodestr)
      (apply 'ee-goto-position pos-spec-list)))

;;;###autoload
(defun find-fline (fname &rest pos-spec-list)
  "Visit a file, and optionally go to a certain position in it.
This function is often used as a hyperlink: if you place the point
at the end of a line like one of those below\n
  # (find-fline \"/etc/resolv.conf\")
  # (find-fline \"~root/\")
  # (find-fline \"$HOME/.emacs\" 25)\n
and type `\\[eval-last-sexp]' Emacs will open the corresponding file.\n
For more information on the format of POS-SPEC-LIST, see the
documentation for the function `ee-goto-position'; for ways to omit part of
the directory specification, see the documentation for the function
`code-c-d'."
  (find-file (substitute-in-file-name fname))
  (apply 'ee-goto-position pos-spec-list))

;;;###autoload
(defun find-fline-nosubst (fname &rest pos-spec-list)
  "Like `find-fline', but without running `substitute-in-file-name' on FNAME."
  (find-file fname)
  (apply 'ee-goto-position pos-spec-list))


;;;;
;;;; eecd  (to ".eecd")
;;;; eecd: change to the current buffer's directory.
;;;;
;;
;; In some modes like info-mode or dired-mode it is not immediately
;; obvious which is the current directory; use "M-x pwd" to learn.
;;
;;;###autoload
(defun eecd (&optional command)
  "Like `eev', but saving a command that will \"cd\" to the current directory.
In most cases this function will get the directory from the
buffer-local variable `current-directory' (just like `pwd' does), but
for W3 buffers visiting local files a hack was introduced to make the
Right Thing happen."
  (interactive)
  (if command (setq command (concat "\n" command)) (setq command ""))
  (if (not (eq major-mode 'w3-mode))
      (eev (concat "cd " default-directory command) nil)
    ;; a hack for when we're visiting a local file in w3-mode
    (let ((url (url-view-url 0)))
      (if (string-match "^file:\\(.*/\\)[^/]*$" url)
	  (eev (concat "cd " (match-string 1 url)) nil)
	(error "Current url is %S, which is not a local file" url)))))



;;;;
;;;; M-e  (to ".M-e")
;;;; A variant of `eval-last-sexp' that is able to pass an argument.
;;;; (find-elnode "Scope")
;;;;

(defun end-of-line-then-eval-last-sexp (ee-arg)
  "Go to the end the line, then evaluate the sexp before point.
The argument to this command is made available to the sexp as the
variable `ee-arg', and some hyperlink functions will behave
differently is this variable is non-nil; `find-___w3', for example,
will call lynx or another browser instead of W3. See the source code
of eev.el for details.\n
This function is usually bound to a key by `ee-invade-global-keymap'."
  (interactive "P")
  (end-of-line)
  (eval-last-sexp nil))



;;;;
;;;; find-w3  (to ".find-w3")
;;;;

(defun find-w3 (localurl &rest pos-spec-list)
  "Like `find-fline', but will open LOCALURL using W3 or other browser.
If the value of `ee-arg' is 1 then use Lynx (via `eev') instead of W3.
The standard way to set `ee-arg' is to execute this function with
`\\[end-of-line-then-eval-last-sexp]' instead of with `\\[eval-last-sexp]'.\n
It is easy to add support for more browsers to this function by
redefining it at your .emacs with a richer `cond'. :-)"
  (cond ((eq ee-arg 1)
	 (ees (format "lynx %s\n" localurl) nil))
	(t (w3-open-local localurl)
	   (apply 'ee-goto-position pos-spec-list))))






;;;;
;;;; code-c-d  (to ".code-c-d")
;;;; code-c-d, a factory of hyperlink functions
;;;;
;;
;; (find-efile "etags.el")
'(load-library "etags")
(require 'etags)			; we need to set tags-add-tables to nil
(setq tags-add-tables nil)		; for find-___tags

;; TO DO: check if we should use (let (tags-add-tables) ...) instead.

(defun format-and-eval (formatstr &rest rest)
  "Apply `format' to FORMATSTR and REST and evaluate the result."
  (eval (read (apply 'format (concat "(progn " formatstr ")") rest))))

(defun find-node2 (infofile nodename &optional pos-spec-list  code)
  "Like `find-node', but saves CODE in `ee-temp-code' (to let `inn' use it)."
  (if code (setq ee-temp-code code ee-temp-infofile infofile))
  (apply 'find-node (format "(%s)%s" infofile nodename) pos-spec-list))

;;;###autoload
(defun code-c-d (c d &optional infofile)
  "Define an entire family of hyperlink functions at once.
A call to `code-c-d' with \"xxx\" as the value of C will define the
hyperlink functions `find-xxxfile', `find-xxxtag', `find-xxxw3' and
optionally `find-xxxnode', plus the auxiliary functions and variables
`ee-xxxdir', `ee-xxxtagsfile', `ee-xxxfile' and `ee-use-xxx-tags'.\n
To best way to understand this function is to make a call to
`icode-c-d' with the same parameters to see the code that `code-c-d'
will execute.  To see an example, copy the two lines below to a
non-read-only buffer and execute each of them with \\[eval-last-sexp]:\n
  # (icode-c-d \"grub\" \"/usr/src/grub-0.5.95/\")
  # (icode-c-d \"grub\" \"/usr/src/grub-0.5.95/\" \"grub\")\n
For more information (examples, tutorials, etc) see
<http://angg.twu.net/>."
  (format-and-eval "
    (setq ee-%sdir \"%s\")
    (setq ee-%stagsfile \"%sTAGS\")
    (defun ee-%sfile (str)
      (concat (substitute-in-file-name ee-%sdir) str))
    (defun ee-use-%s-tags ()
      (setq tags-file-name ee-%stagsfile))
    (defun find-%sfile (str &rest pos-spec-list)
      (ee-use-%s-tags)
      (apply 'find-fline-nosubst (ee-%sfile str) pos-spec-list))
    (defun find-%stag (str)
      (ee-use-%s-tags) (find-tag str))
    (defun find-%sw3 (furl &rest pos-spec-list)
      (apply 'find-w3 (ee-%sfile furl) pos-spec-list))
    (setq ee-temp-code %S ee-temp-infofile %S)
    " c d  c d  c c  c c  c c c  c c  c c  c infofile)
  (if infofile
     (format-and-eval "
       (defun find-%snode (nodename &rest pos-spec-list)
          (find-node2 %S nodename pos-spec-list  %S))" c infofile  c)))

;; code-c-d-explan1  (to ".code-c-d-explan1")
;;
;; To understand exactly what a given "code-c-d" does, put an "i"
;; before the "code-c-d" and run it; the function "icode-c-d", defined
;; below, is a hack that will insert at point the block of code that
;; code-c-d would eval. For example,
;;
;;   (icode-c-d "awk" "/usr/src/gawk-3.0.3/" "gawk")
;;
;; gives, modulo indentation,
;;
;; (progn
;;   (setq ee-awkdir "/usr/src/gawk-3.0.3/")
;;   (setq ee-awktagsfile "/usr/src/gawk-3.0.3/TAGS")
;;   (defun ee-awkfile (str)
;;     (concat (substitute-in-file-name ee-awkdir) str))
;;   (defun ee-use-awk-tags ()
;;     (setq tags-file-name ee-awktagsfile))
;;   (defun find-awkfile (str &rest pos-spec-list)
;;     (ee-use-awk-tags)
;;     (apply 'find-fline-nosubst (ee-awkfile str) pos-spec-list))
;;   (defun find-awktag (str)
;;     (ee-use-awk-tags) (find-tag str))
;;   (defun find-awkw3 (furl &rest pos-spec-list)
;;     (apply 'find-w3 (ee-awkfile furl) pos-spec-list))
;;   (setq ee-temp-code "awk" ee-temp-infofile "gawk"))
;;
;; (progn
;;   (defun find-awknode (nodename &rest pos-spec-list)
;;     (find-node2 "gawk" nodename pos-spec-list  "awk")))
;;
;; where the second progn block would not have been generated if the
;; third argument to code-c-d (the name of an info file) had been
;; suppressed.
;;
;; code-c-d-explan2  (to ".code-c-d-explan2")
;;
;; That is, the first argument to a code-c-d is the "prefix", the
;; second if the "path", the third is the "info file". A call to
;; code-c-d with a prefix of "xxx" defines ee-xxxdir, ee-xxxtagsfile,
;; ee-xxxfile, ee-use-xxx-tags, find-xxxfile, find-xxxtag, find-xxxw3
;; and maybe find-xxxnode; there are lots of examples of usage of
;; these functions in this file and in my e-scripts directory, which
;; is at <http://angg.twu.net/escripts.html>.

;; This is a quick hack that only serves dydactical purposes.
;;
(defun icode-c-d (&rest args)
  "A variation of `code-c-d' that will insert the generated code at point instead of executing it."
  (defun format-and-eval (formatstr &rest rest)
    (insert (apply 'format (concat "\n(progn " formatstr ")\n") rest)))
  (apply 'code-c-d args)
  (defun format-and-eval (formatstr &rest rest)
    (eval (read (apply 'format (concat "(progn " formatstr ")") rest)))))



;;;;
;;;; find-anchor  (to ".find-anchor")
;;;; Find strings enclosed in angle quotation marks.
;;;; These functions are especially convenient for when you have to
;;;; convert your files with hyperlinks to html.
;;;;
;;
(defun find-anchor (fname &optional tag &rest rest)
  "An auxiliary function to find \"anchor strings\" in text files."
  (find-fline fname)
  (if tag (apply 'ee-goto-position (format ee-anchor-format tag) rest)))

;; find-Package  (to ".find-Package")
;; For info on Debian packages. See `dff' below.
;; Here we use another format for anchors.
;;
(defun find-Package (fname &optional packagename &rest rest)
  "An auxiliary function to find package descriptions in Debian \"Packages\" files."
  (find-fline fname)
  (if packagename
      (apply 'ee-goto-position (format "\nPackage: %s\n" packagename) rest)))

(defun find-status (packagename &rest rest)
  "A hyperlink to the information in /var/lib/dpkg/status about a Debian package.
See the documentation on `dff'."
  (apply 'find-Package "/var/lib/dpkg/status" packagename rest))

(defun find-available (packagename &rest rest)
  "A hyperlink to the information in /var/lib/dpkg/available about a Debian package.
See the documentation on `dff'."
  (apply 'find-Package "/var/lib/dpkg/available" packagename rest))




;; find-efunction  (to ".find-efunction")
;; find-evariable  (to ".find-evariable")

(defun find-efunction (symbol &rest rest)
  (let ((location (find-function-noselect symbol)))
    (switch-to-buffer (car location))
    (goto-char (cdr location))
    (ee-goto-rest rest)))

(defun find-evariable (symbol &rest rest)
  (let ((location (find-variable-noselect symbol)))
    (switch-to-buffer (car location))
    (goto-char (cdr location))
    (ee-goto-rest rest)))

;; (find-efunction 'view-lossage)
;; (find-efunction 'code-c-d)
;; (find-efunction 'find-efile)

;; find-progoutput  (to ".find-progoutput")
;; Experimental, and needs a better name.
;;
;; (find-elnode "Synchronous Processes" "Function: shell-command-to-string")
;; (find-elnode "Buffer Basics")
;; (find-elnode "Buffer Names" "Function: get-buffer")
;; (find-progoutput "wget --help" "--continue")
;; (find-progoutput "set")

(defun find-progoutput (command &rest pos-spec-list)
  (if (get-buffer command)		; if the buffer already exists
      (switch-to-buffer command)	; then just switch to it
    (switch-to-buffer command)		; otherwise create it
    (insert (shell-command-to-string command)) ; prepare its contents
    (goto-char (point-min)))		; and place point at its beginning
  (apply 'ee-goto-position pos-spec-list))





;;;;
;;;; to_and_back  (to ".to_and_back")
;;;; Hyperlinks to anchors in the same buffer, with a "back" function.
;;;; The "ee" prefix is somewhat misleading: these functions are
;;;; related to following hyperlinks, not to saving blocks of text.
;;;;
;;
;; Note: the "back" list is currently buffer-local, and saves only line
;; numbers. Reason: this was easier to code.
;; (find-elnode "Intro to Buffer-Local" "`make-variable-buffer-local'")
;;
(make-variable-buffer-local 'ee-back-list)
(setq-default ee-back-list nil)		; for XEmacs

(defun ee-save-position ()
  "Push an `(ee-goto-position LINENOW)' at the head of `ee-back-list'."
  (setq ee-back-list
	(cons (list 'ee-goto-position (count-lines (point-min) (point)))
	      ee-back-list)))

(defun ee-to (anchor)
  (interactive "sAnchor: ")
  "Local hyperlink to ANCHOR, pushing the current line into `ee-back-list'."
  (ee-save-position)
  (ee-goto-position (format ee-anchor-format anchor)))

(defun ee-back ()
  "Pop `ee-back-list', going to the position stored in the popped element."
  (interactive)
  (eval (car ee-back-list))
  (setq ee-back-list (cdr ee-back-list)))




;; find-man  (to ".find-man")
;; (find-eev "eev-hybrids.el" "find-man.el")
;; (find-eev "eev-hybrids.el" "eeman")
;;
(defun find-man (manpage &rest pos-spec-list)
  (if (null ee-arg) (apply 'find-man.el manpage pos-spec-list)
    (apply 'eeman manpage pos-spec-list)))

;; (find-man.el "1 more" "-num")



;; Local Variables:
;; coding:               no-conversion
;; ee-anchor-format:     "%s"
;; ee-charset-indicator: ""
;; End:

;;; eev-links.el ends here
