;; Copyright (C) 1999, 2000, 2001, 2002 Free Software Foundation, Inc.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2002may11
;; Keywords:   escripts, hyperlinks, shell, tcl, perl, tex
;;
;; This file was copylefted to prevent against patent psychopaths; if
;; you want a version with any other license you'll have to write it
;; yourself. More formally,
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;;
;; URLs for this file:
;;   http://angg.twu.net/eev/eev-mklinks.el
;;   http://angg.twu.net/eev/eev-mklinks.el.html

;;; Commentary:

;; This package provides support for e-scripts in Emacs. Look in
;; <http://angg.twu.net/emacs.html> for more information and in
;; <http://angg.twu.net/escripts.html> for examples of e-scripts... an
;; e-script is typically a text file in which certain parts are meant
;; to be executed by the shell or by other programs and certain
;; comments are in fact Emacs Lisp code; executing one of these
;; comments with C-x C-e (i.e., eval-last-sexp) brings up a file or an
;; info node, and thus works as a hyperlink. Many of the comments in
;; this file can be executed in this way.

;; See also the "eev Manifesto", at
;; <http://angg.twu.net/eev-manifesto.html>, the section named
;; "eev-howto", below, and whatever else seem important in
;; <http://angg.twu.net/>; that site is almost completely dedicated to
;; using eev.el.

;;;;
;;;; ee-insert-info-hyperlink
;;;; Also called:         inn  (to ".inn")
;;;; Insert a hyperlink to the current info node.
;;;; 
;;
;; (find-elnode "Interactive Examples")
;; (find-efile "info.el")
;; (find-efile "info.el" "Convert filename to lower case")
;; (find-etag "Info-set-mode-line")
;;
(defun ee-infofile-to-code (infofile)
  "Try to convert INFOFILE to a \"code\" that can be used by `inn'.
This function is quite primitive at present: it just takes the
non-directory parts of INFOFILE and of `ee-temp-infofile', and if they
are equal it returns `ee-temp-code' (else it returns nil)."
  (if (and infofile
	   ee-temp-infofile
	   (string= (file-name-nondirectory infofile)
		    (file-name-nondirectory ee-temp-infofile)))
      ee-temp-code))

(defun ee-insert-info-hyperlink (arg)
  "Insert a hyperlink to the current info node.
If ARG is non-nil try to use the short form, like `# (find-elnode \"Top\")'.
Without an argument always use long forms, like `(find-node \"(elisp)Top\")'.
The infofile part (\"(elisp)\" in the example) is obtained by just taking
the non-directory part of `Info-current-file', and thus may be incomplete
if the infofile was at a non-standard directory; use `code-c-d' to work
around that."
  (interactive "P")
  (let (infofile infonode f code s)
    (if (get-buffer "*info*")
	(progn
	  (save-excursion
	    (set-buffer "*info*")
	    (setq infofile Info-current-file
		  infonode Info-current-node
		  f (file-name-nondirectory infofile)))
	  (if arg (setq code (ee-infofile-to-code infofile)))
	  ;; (insert (format "dbg: %S %S %S\n" infofile ee-temp-infofile f))
	  (insert
	   (format ee-comment-format
		   (if code (format "(find-%snode \"%s\")" code infonode)
		     (format "(find-node \"(%s)%s\")" f infonode)))))
      (error "No curent info node"))))


;;;;
;;;; ee-convert-to-debian-link
;;;; Also called:          dff  (to ".dff")
;;;; This is a macro to create find-xxx hyperlinks for debian packages.
;;;;
;;
;; (find-elnode "Keyboard Macros")
;;
(defun ee-convert-to-debian-link (N)
  "Convert N names of Debian packages into hyperlinks to the packages.
For example, if you have two lines like:\n
  grub
  oskit\n
in a non-read-only buffer and the point is on the first one, then
running `M-2 M-x dff' will convert them to:\n
  # (find-status \"grub\")
  # (find-vldifile \"grub.list\")
  # (find-fline \"/usr/doc/grub/\")
  # (find-status \"oskit\")
  # (find-vldifile \"oskit.list\")
  # (find-fline \"/usr/doc/oskit/\")\n"
  (interactive "p")
  (execute-kbd-macro
   (read-kbd-macro "C-a NUL C-e C-w
                    # SPC (find-status SPC \" C-y \") RET
                    # SPC (find-vldifile SPC \" C-y .list\") RET
                    # SPC (find-fline SPC \"/usr/doc/ C-y /\") C-a <down>")
   N))



;;; eev-mklinks.el ends here
