;;; eev-save.el --- embed shell/TeX/etc code in plain text.

;; Copyright (C) 1999, 2000, 2001, 2002 Free Software Foundation, Inc.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2002may11
;; Keywords:   escripts, hyperlinks, shell, tcl, perl, tex
;;
;; This file was copylefted to prevent against patent psychopaths; if
;; you want a version with any other license you'll have to write it
;; yourself. More formally,
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.
;;
;; URLs for this file:
;;   http://angg.twu.net/eev/eev-save.el
;;   http://angg.twu.net/eev/eev-save.el.html

;;; Commentary:

;; This package provides support for e-scripts in Emacs. Look in
;; <http://angg.twu.net/emacs.html> for more information and in
;; <http://angg.twu.net/escripts.html> for examples of e-scripts... an
;; e-script is typically a text file in which certain parts are meant
;; to be executed by the shell or by other programs and certain
;; comments are in fact Emacs Lisp code; executing one of these
;; comments with C-x C-e (i.e., eval-last-sexp) brings up a file or an
;; info node, and thus works as a hyperlink. Many of the comments in
;; this file can be executed in this way.

;; See also the "eev Manifesto", at
;; <http://angg.twu.net/eev-manifesto.html>, the section named
;; "eev-howto", below, and whatever else seem important in
;; <http://angg.twu.net/>; that site is almost completely dedicated to
;; using eev.el.

;;;;
;;;; eev_pre  (to ".eev_pre")
;;;; Some basic functions for saving blocks of text as temporary scripts.
;;;;
;;
;; (find-elnode "Writing to Files" "If START is a string")
;; (find-elnode "File Name Expansion" "substitute-in-file-name")
;; s: "start of region", or "string";
;; e: "end of region".

(defun ee-se-to-string (s e)
  "Return the part of the current buffer between S and E, or S if S is a string."
  (cond ((numberp s) (buffer-substring s e))
	((stringp s) s)))

(defun octal-to-num (str)
  "Interpret STR as an octal number and return the result."
  (let ((lastv (- (string-to-char (substring str -1)) ?0))
	(rest (substring str 0 -1)))
    (if (string= "" rest) lastv (+ lastv (* 8 (octal-to-num rest))))))

(defun ee-write-string (str &optional other-ee-file fmode)
  "Write STR into the specified file and optionally set the file's permissions.
If OTHER-EE-FILE is nil, use the value of the variable `ee-file' as
the file name; if not, use OTHER-EE-FILE.  In both cases
`substitute-in-file-name' will be applied to the file name, to make
things like \"$HURDDIR/root/bin/ee.sh\" be interpreted in the right
way.\n
If FMODE is given then it will be converted to a number by
`octal-to-num' and used to set the permissions of the file.\n
Note that if ange-ftp is not disabled and the resulting filename is
something of the form \"/HOST:FILENAME\" or \"/USER@HOST:FILENAME\"
then Emacs will call an FTP program to write the string to a file on a
remote machine."
  (let ((fname (substitute-in-file-name (or other-ee-file ee-file))))
    (write-region str nil fname)	; a standard kludge
    (if fmode (set-file-modes fname (octal-to-num fmode)))))

(defun ee-write (s e pre post &optional other-ee-file fmode)
  "Same as `ee-write-string', but build the string from S, E, PRE and POST."
  (ee-write-string (concat pre (ee-se-to-string s e) post)
		   other-ee-file fmode))


;; Some shells won't display the last command of an eev-ed block in a
;; nice way if the last line of it doesn't end with a newline, so we
;; need this kludge... note that `eev' calls `ee-write-with-nl'
;; instead of `ee-write'.
;;
;; (find-elnode "Regexp Search" "Function: string-match")
;; (find-elnode "Syntax of Regexps" "`\\''")
;; (string-match "[^\n]\\'" "")
;; (string-match "[^\n]\\'" "A line that terminates correctly\n")
;; (string-match "[^\n]\\'" "Cases like this need an extra newline")

(defun ee-se-to-string-with-nl (s e)
  "Same as `ee-se-to-string',\
 but forces the last line of the string to end with a newline."
  (let ((str (ee-se-to-string s e)))
    (if (string-match "[^\n]\\'" str) (concat str "\n") str)))

(defun ee-write-with-nl (s e pre post &optional other-ee-file fmode)
  "Same as `ee-write',\
 but forces the last line of the string to end with a newline."
  (ee-write-string (concat pre (ee-se-to-string-with-nl s e) post)
		   other-ee-file fmode))





;;;;
;;;; eev  (to ".eev")
;;;;

;;;###autoload
(defun eev (s &optional e other-ee-file)
  "Save the region (plus \"set -v\"/\"set +v\") to a temporary script file.
If you have configured your shell properly (see below) then the shell
command \"ee\" will run the lines from the region similarly to as if
they were being typed at the keyboard -- each command will be echoed
before being executed, and commands like \"cd\", \"set\" and \"alias\" will
keep their effects after the script finishes.\n
See the documentation on `ee-se-to-string' on how S and E are
converted to the contents of the region. The name of the temporary
script file is taken from OTHER-EE-FILE if it is non-nil, else from
`ee-file'.\n
Configuration, and a simple example: if your shell is bash or zsh then
configuring it for eev means adding the following lines to your
.bashrc/.zshrc:\n
  alias ee='. ~/bin/ee'
  export EE=~/bin/ee.sh
  export EEG=~/bin/ee.generic\n
Once this is done you can set the region to the block below and then
run `M-x eev' in Emacs and \"ee\" in a shell; the result should be easy
to guess.\n
# (find-fline \"~/eev.el\" \"hyperlinks are comments and are ignored\")
cd
pwd
cd /tmp/
cat <<'---'
  Hi
---
pwd"
  (interactive "r")
  (ee-write-with-nl s e "" "" other-ee-file))

;;;###autoload
(defun eevt (s &optional e other-ee-file)
  "Save the region to a temporary script file (tcsh version).
See the comments in \"eev.el\"."
  (interactive "r")
  (ee-write s e "set verbose\n" "\nunset verbose\n" other-ee-file)) ; tcsh

;; eex: uses "-x" instead of "-v"
;; ees: "silent", i.e., non-verbose
;;
(defun eex (s e)
  "A variant of `eev' that uses \"set -x\"/\"set +x\" instead of \"set -v\"/\"set +v\"."
  (interactive "r")
  (ee-write s e "set -x\n" "\nset +x"))

(defun ees (s e)
  "\"eev-silent\": like `eev', but doesn't set up any kind of verbose mode."
  (interactive "r")
  (ee-write s e "" "\n"))

;; eegdb: save a block as a "gdb script".
;;
(defun eegdb (s &optional e)
  "Save the region to the temporary \"gdb script\" file.
See the comments in \"eev.el\"."
  (interactive "r")
  (ee-write s e "" "" ee-file-gdb))



;;;;
;;;; eeg  (to ".eeg")
;;;; The "generic" version of eev.
;;
;; The main use of this function is in connection with an Expect
;; script that is also called "eeg", and that can be found at:
;; <http://angg.twu.net/EXPECT/eeg>
;; (find-angg "EXPECT/eeg")
;;
(defun eeg (s e)
  "Like `eev', but for the \"generic interface\".
Technically, this function just saves the region to the file whose
name is given by the value of `ee-file-generic' -- but you can use a
simple Expect script to paste lines from that file to any application
that takes its input from stdin, and so this gives you a sort of `eev'
that will work on programs where you can't implement an \"ee\"
command.\n
The Expect script is at <http://angg.twu.net/EXPECT/eeg>; for more
information take a look at its comments and at the comments in the
file \"eev.el\".\n"
  (interactive "r")
  (ee-write s e "" "" ee-file-generic))



;;;;
;;;; eelatex  (to ".eelatex")
;;;; A function to run LaTeX on a block of text.
;;;;
;;
(defun eelatex (s e)
  "An `eev'-like function to run LaTeX on a block of text.
This function will write the string obtained by concatenating
`eelatex-pre', the region, and `eelatex-post' to the temporary LaTeX
file whose name is given by the variable `eelatex-file', and will write
the commands in the variable `eelatex-eevscript' to the temporary
script file given by `ee-file' (the same that is used by `eev'). The
effect of all that is that by going to a properly-configured shell and
typing \"ee\" the temporary LaTeX file will be LaTeX-ed.\n
Note that you can put a \"Local Variables:\" section at the end of a
LaTeX file to make this command use special settings to compile pieces
of the file. For example:\n
  % Local Variables:
  % eelatex-file: \"~/LATEX/tmp.tex\"
  % eelatex-pre: \"\\input mythesis.sty\\n\\begin{document}\\n\"
  % eelatex-eevscript: \"cd ~/LATEX/; make tmp.dvi\"
  % End:"
  (interactive "r")
  (ee-write s e eelatex-pre eelatex-post eelatex-file)
  (eev eelatex-eevscript nil)
  (message "eelatex: wrote %s and %s" eelatex-file ee-file))




;;;;
;;;; bounded  (to ".bounded")
;;;; "Bounded" versions for some eev commands; instead of working
;;;; with the region these functions select the text around point
;;;; until some fixed delimiter strings.
;;;;
;;;; Note that `eev-bounded', `eelatex-bounded' and `eeg-bounded' set
;;;; the variable `ee-temp-bounded-function' to their function symbol
;;;; when they are run and this lets `ee-bounded' re-execute the last
;;;; eexxx-bounded command used.
;;;;
;;
;; bounded_support  (to ".bounded_support")
;; Some support functions,
;; (find-elnode "String Search")
;;
(defun ee-search-backward (str)
  "Return the position after the previous occurrence of STR."
  (+ (save-excursion (search-backward str))
     (length str)))
(defun ee-search-forward (str)
  "Return the position before the next occurrence of STR."
  (- (save-excursion (search-forward str))
     (length str)))

(defun ee-bounded ()
  "Re-execute the last `eexxx-bounded' function called.
If you think you are going to use some of the `eexxx-bounded'
functions -- `eev-bounded' or `eelatex-bounded', say -- frequently
then this function certainly deserves to be bound to a key. See
`ee-set-keymap'."
  (interactive)
  (funcall ee-temp-bounded-function))

(defun ee-meta-bounded (boundedfun fun sstr estr &rest rest)
  "A function that helps building `eexxx-bounded' functions.
It sets `ee-temp-bounded-function' to BOUNDEDFUN (for `ee-bounded')
and runs FUN with the string between SSTR and ESTR and extra arguments
REST.\n
If BOUNDEDFUN is nil then `ee-temp-bounded-function' is not set."
  (if boundedfun (setq ee-temp-bounded-function boundedfun))
  (apply fun
	 (ee-search-backward sstr)
	 (ee-search-forward estr)
	 rest))

;; eexxx-bounded  (to ".eexxx-bounded")
;; and the `eexxx-bounded' themselves:
;;
(defun eev-bounded (&optional once)
  "Run `eev' on the region delimited by `ee-delimiter-hash'es.
Here's a simple example.  First execute the S-expression below, just
to make sure that the delimiter string is set to the right value:\n
  (set (make-local-variable 'ee-delimiter-hash) \"\\n#-\\n\")\n
Now place the cursor somewhere between the \"#-\"s below and run
`eev-bounded'; then go to a shell that was configured for `eev' (see
the documentation for the function `eev') and run \"ee\" as usual.\n
#-
cd
pwd
# Hyperlink comments are allowed as usual:
# (find-node \"(elisp)Creating Buffer-Local\")
# (find-node \"(bash)Redirections\" \"Here Documents\")
cat <<'---'
  Hello!
---
cd /tmp
#-\n
See also the documentation on `\\[ee-default-strbounded]'."
  (interactive)
  (ee-meta-bounded (if once nil 'eev-bounded)
		   'eev ee-delimiter-hash ee-delimiter-hash))

(defun eelatex-bounded (&optional once)
  "Run `eelatex' on the region delimited by `ee-delimiter-percent's.
See the documentation on `\\[ee-default-strbounded]'."
  (interactive)
  (ee-meta-bounded (if once nil 'eelatex-bounded)
		   'eelatex ee-delimiter-percent ee-delimiter-percent))

(defun eeg-bounded (&optional once)
  "Run `eeg' on the region delimited by `ee-delimiter-hash'es.
See the documentation on `\\[ee-default-strbounded]'."
  (interactive)
  (ee-meta-bounded (if once nil 'eeg-bounded)
		   'ee-write ee-delimiter-hash ee-delimiter-hash
		   "" "" ee-file-generic))

(defun eegdb-bounded (&optional once)
  "Run `eegdb' on the region delimited by `ee-delimiter-hash'es.
See the documentation on `\\[ee-default-strbounded]'."
  (interactive)
  (ee-meta-bounded (if once nil 'eegdb-bounded)
		   'eegdb ee-delimiter-hash ee-delimiter-hash))



;; Local Variables:
;; coding:               no-conversion
;; ee-anchor-format:     "%s"
;; ee-charset-indicator: ""
;; End:

;;; eev-save.el ends here
