;;; eev.el --- embed shell/tcl/perl code and elisp hyperlinks in plain text.
;; Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.

;; Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
;; Maintainer: Eduardo Ochs <edrx@mat.puc-rio.br>
;; Version:    2001dec26
;; Keywords:   escripts, hyperlinks, shell, tcl, perl, tex
;;
;; This file was copylefted to prevent against patent psychopaths; if
;; you want a version with any other license you'll have to write it
;; yourself. More formally,
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.

;;; Commentary:

;; This package provides support for e-scripts in Emacs. Look in
;; <http://angg.twu.net/emacs.html> for more information and in
;; <http://angg.twu.net/escripts.html> for examples of e-scripts... an
;; e-script is typically a text file in which certain parts are meant
;; to be executed by the shell or by other programs and certain
;; comments are in fact Emacs Lisp code; executing one of these
;; comments with C-x C-e (i.e., eval-last-sexp) brings up a file or an
;; info node, and thus works as a hyperlink. Many of the comments in
;; this file can be executed in this way.

;; See also the "eev Manifesto", at
;; <http://angg.twu.net/eev-manifesto.html>, the section named
;; "eev-howto", below, and whatever else seem important in
;; <http://angg.twu.net/>; that site is almost completely dedicated to
;; using eev.el.

;; Here's an index for the main sections of this file.
;; If you have ran the two `ee-invade-global-xxx' functions then you
;; should be able to follow the "to" hyperlinks with `M-e'.
;;
;; Basic building blocks:
;; .find-		(to "find-")
;; .variables		(to "variables")
;; .eev_pre		(to "eev_pre")
;;
;; A way to pass a parameter to hyperlink functions:
;; .ee-arg		(to "ee-arg")
;; .M-e		(to "M-e")
;;
;; The `eev' function and its closest friends:
;; .eev-explanation	(to "eev-explanation")
;; .eev-howto		(to "eev-howto")
;; .eev-bash		(to "eev-bash")
;; .eev-tcsh		(to "eev-tcsh")
;; .eev		(to "eev")
;; .eeg		(to "eeg")
;;
;; Other, much stranger, `eexxx' functions:
;; .eecd		(to "eecd")
;; .eelatex		(to "eelatex")
;;
;; `eev-bounded' and many functions like it:
;; .bounded		(to "bounded")
;; .bounded_support	(to "bounded_support")
;; .eexxx-bounded	(to "eexxx-bounded")
;;
;; .find-w3		(to "find-w3")
;;
;; A single call to `code-c-d' lets you define many `find-xxx'
;; hyperlink functions at once.
;; .code-c-d		(to "code-c-d")
;; .code-c-d-explan1	(to "code-c-d-explan1")
;; .code-c-d-explan2	(to "code-c-d-explan2")
;; .code-c-d-examples	(to "code-c-d-examples")
;;
;; Strings within certain delimiters can be treated as "anchors",
;; i.e., special targets for `find-' functions.
;; .find-anchor	(to "find-anchor")
;; .find-Package	(to "find-Package")
;; .to_and_back	(to "to_and_back")
;;
;; Two functions to create hyperlinks.
;; .inn		(to "inn")
;; .dff		(to "dff")
;;
;; Many examples of things that can be done with `eexxx' functions.
;; .eexxx_examples	(to "eexxx_examples")
;;
;; To customize Emacs to make it more eev-friendly.
;; .ee-invade-global-	(to "ee-invade-global-")

;; Misc and hybrids:
;; .eeman		(to "eeman")
;; .find-man.el	(to "find-man.el")
;; .find-man		(to "find-man")
;; .find-efunction	(to "find-efunction")
;; .find-evariable	(to "find-evariable")
;; .menu-bar		(to "menu-bar")
;; .ee-load-back	(to "ee-load-back")
;; .find-progoutput	(to "find-progoutput")

;;; History:

;; 2001jul11: added `find-available'.
;; 2001jan26: renamed `inn' and `dff', added `ee-to', `ee-back' and
;;            the `ee-invade-global-' functions; added an index and
;;            rewrote/rearranged great part of the comments.
;; 2001jan14: changed the last codepage850 chars in comments to latin1.
;; 2001jan07: added the optional argument `once' to bounded functions.
;; 2000oct30: changed `eelatex' to make it use the new variables
;;            `eelatex-post', `eelatex-file' and `eelatex-eevscript'.
;;            `inn' can now generate short info links for infofiles
;;            outside the standard info path; the long form for these
;;            links still doesn't include enough of the path.
;; 2000oct24: `ee-temp-bounded-function' is now marked as a
;;            "safe-local-variable" and thus it can now be set in the
;;            "Local Variables:" section of a file.
;; 2000sep21: fixed two little bugs in `find-anchor' and `inn'.
;; 2000sep08: added `ee-comment-format'.
;; 2000sep06: renamed `ee-default-bounded-function' to
;;            `ee-temp-bounded-function', cleaned the code that used
;;            it, and did some other renamings; now the variables that
;;            are set by `find-node2' and used by `inn' are called
;;            `ee-temp-infofile' and `ee-temp-code', for example.
;; 2000aug28: many more docstrings.
;; 2000aug22: renamed `find-tagh' to `find-anchor'; added
;;            `ee-anchor-string', `ee-delimiter-hash',
;;            `ee-delimiter-percent' and changed several functions to
;;            use them; added some more docstrings.
;; 2000aug04: corrected the instructions for tcsh; added the prefix
;;            `ee-' to some variables and internal functions created
;;            by `code-c-d'.
;; 2000jul04: added `find-w3' (finally!) and `ee-arg'.
;; 2000jun29: added `end-of-line-then-eval-last-sexp' and lots of
;;            comments.
;; 2000jun20: added `ee-default-bounded-function' and its friends.
;;            Added many `code-c-d's for Debian 2.2.
;; 2000may27: `find-___w3' now supports pos-specs, and `eecd' will now
;;            work in w3-mode.
;; 2000may26: added `find-status' and changed `dff' to use it.
;; 2000may25: Fixed some bugs in `inn' (but it will still require the
;;            exact case in the code-c-d to work right with an
;;            argument).
;; 2000may18: Added `find-fline-nosubst' and changed `code-c-d'; now
;;            the non-prefix part of a filename can have "~"s without
;;            them getting, huh, "expanded".
;; 2000may10: The `find-___' functions now accept any number of
;;            `pos-spec's. The modified `goto-position' is ugly and
;;            huge, so I haven't included it here; you'll have to
;;            fetch it from http://angg.twu.net/ and put it in your
;;            .emacs if you want to use it.
;; 2000may09: The generic interface. Removed all references to the
;;            `EE' environment variable (that used to hold the value
;;            of `ee-file'. The gdb interface now uses
;;            ee-file-generic.
;; 2000mar11: Got a review in Brave GNU World #13! (And maybe my first
;;            serious users... 8-)
;; 2000feb12: First announcement on FreshMeat.
;; late 1999: I started to have the impression that using elisp code
;;            for hyperlinks was not common practice.  Got in touch
;;            with the GNU people and they told me to start writing
;;            documentation.  Split eev.el from my .emacs.
;; early '99: `eeman', bounded `ee's, `inn', `dff'.  Started my
;;            e-scripts repository, at http://mat.puc-rio.br/~edrx/.
;;    1995-6: First functions: `find-fline', `eev', and then
;;            `find-node' and `code-c-d'.

;;; Code:

;;;;
;;;; eev-explanation  (to ".eev-explanation")
;;;; eev and friends.
;;;;
;;
;; "M-x eev" writes the "region" of the current buffer (i.e.,
;; everything between the cursor and the "mark") to a file, so that it
;; can be executed as a script. In fact it does better than that: the
;; region is saved surrounded by "set -v" and "set +v" and we source
;; it instead of running it as a script, to let it use and change the
;; environment, the directory and the pool of aliases and functions of
;; the current shell. Marking a block of text in emacs, typing "C-x
;; eev" and then going to a shell and running "ee" (which should be
;; aliased to `. ~/bin/ee.sh') is almost 100% equivalent to typing
;; each line to the shell.
;;
;; eev-howto  (to ".eev-howto")
;; That was too technical... Here's the "quick eev how-to": copy
;; eev.el to your home directory, put these three lines in your .emacs
;; and restart Emacs,
;;
;;   (load-library "~/eev.el")
;;   (ee-invade-global-namespace)
;;   (ee-invade-global-keymap)
;;
;; or just execute them with C-x C-e if you already understand how; in
;; the latter case you won't need to restart Emacs. Also, put the
;; following line in your .zshrc (or .bashrc) and start a new
;; bash/zsh, or just execute it manually; the important thing is to
;; make this alias active.
;;
;;   alias ee='. ~/bin/ee.sh'
;;
;; Now write some shell commands in Emacs; mark a block of text around
;; them and run M-x eev. You should get a message like "wrote
;; ~/bin/ee.sh" in the echo area at the bottom of the screen. Switch
;; to a zsh/bash in which this alias is active (probably in another
;; virtual console or in another window) and run "ee". Voila`! Your
;; commands were executed, and each of them was be displayed (without
;; pause, and without a shell prompt) before being executed.
;;
;; eev-bash  (to ".eev-bash")
;; Note: in bash the last line ("set +v") doesn't get overwritten by
;; the next shell prompt and remains on the screen. (Ugly. I need to
;; fix this, but I don't know how).
;;
;; eev-tcsh  (to ".eev-tcsh")
;; If you use tcsh then things are a bit worse. This is what you
;; should put in your .cshrc:
;;
;;   alias ee 'source ~/bin/ee.sh'
;;
;; and as the commands to put the shell in verbose mode are different
;; in tcsh you should either use eevt instead of eev or, better, make
;; the definition of eevt overwrite the definition of eev with:
;;
;;   (load-library "~/eev.el")
;;   (defalias
;;   (defun eev (s e &optional other-ee-file) (interactive "r")
;;     (ee-write s e "set verbose\n" "\nunset verbose" other-ee-file))
;;
;; See:
;; (find-node "(zsh)Shell Builtin Commands" "`source'")
;; (find-node "(zsh)Shell Builtin Commands" "`set ")
;; (find-node "(zsh)Shell Builtin Commands" "`.")
;; (find-node "(zsh)Description of Options" "-v")
;; (find-node "(bash)Bourne Shell Builtins" "`.'")
;; (find-node "(bash)The Set Builtin" "  `-v'")
;; (find-elnode "Markers")

;;;;
;;;; ee-invade-global-  (to ".ee-invade-global-")
;;;; Functions to customize Emacs for eev in convenient but ugly ways.
;;;; They will pollute your global key map and global namespace a bit.
;;;;
;;
(defun ee-invade-global-namespace ()
  "Define some aliases to hyperlink functions with long names."
  (interactive)
  (defalias 'to 'ee-to)
  (defalias 'back 'ee-back)
  (defalias 'inn 'ee-insert-info-hyperlink)
  (defalias 'dff 'ee-convert-to-debian-link))

(defun ee-invade-global-keymap ()
  "Set some keys in the global to functions related to `eev'."
  (interactive)
  (global-set-key [f3] 'ee-bounded)
  (global-set-key "\M-k" 'kill-buffer)
  (global-set-key "\M-E" 'end-of-line-then-eval-last-sexp)
  (global-set-key "\M-e" 'end-of-line-then-eval-last-sexp)
  (global-set-key "\M-A" 'back))

(defun ee-invade-global-menu-bar ()
  "Add an \"(eev)\" entry to the top-level menu bar."
  (interactive)
  (define-key global-map [menu-bar eev] (cons "(eev)" menu-bar-eev-menu)))

;; (find-eev "demo/.emacs")
;; (find-eev "eev-util" "install_dotfiles")
;; (find-angg ".emacs" "load_eev.el")

(defun eev-install-aggressively ()
  (interactive)
  (ee-invade-global-namespace)
  (ee-invade-global-keymap)
  (ee-invade-global-menu-bar)
  (ee-install-find-man.el-hack))

(defun eev-load-libraries ()
  (interactive)
  (load-library "eev-vars")
  (load-library "eev-links")
  (load-library "eev-mklinks")
  (load-library "eev-save")
  (load-library "eev-misc")
  (load-library "eev-hybrids")
  (load-library "eev-menu-bar")
  (load-library "eev-gud"))





;; Local Variables:
;; coding:               no-conversion
;; ee-anchor-format:     "%s"
;; ee-charset-indicator: ""
;; End:

;;; eev.el ends here
